/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "operator/operator.pmh"

#define	FE_NOP_DEBUG	FALSE

namespace fe
{

void SurfaceNormalOp::initialize(void)
{
	catalog<String>("icon")="FE_beta";

	catalog<String>("Group");
	catalog<String>("Group","label")="Point Group";
	catalog<String>("Group","suggest")="pointGroups";

	catalog<bool>("PointNormals")=true;
	catalog<String>("PointNormals","label")="Add Point Normals";

	catalog<bool>("PrimitiveNormals")=false;
	catalog<String>("PrimitiveNormals","label")="Add Primitive Normals";

	catalog< sp<Component> >("Input Surface");
}

void SurfaceNormalOp::handle(Record& a_rSignal)
{
	const BWORD pointNormals=catalog<bool>("PointNormals");
	const BWORD primitiveNormals=catalog<bool>("PrimitiveNormals");

	if(!pointNormals && !primitiveNormals)
	{
		return;
	}

	const String group=catalog<String>("Group");
	const BWORD grouped=(!group.empty());

	sp<SurfaceAccessibleI> spSurfaceOutput;
	if(!accessOutput(spSurfaceOutput,a_rSignal)) return;

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,spSurfaceOutput,e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,spSurfaceOutput,e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputGroup;
	if(grouped)
	{
		if(!access(spOutputGroup,spSurfaceOutput,e_pointGroup,group)) return;
	}
	else
	{
		spOutputGroup=spOutputPoint;
	}

	const U32 pointCount=spOutputPoint->count();

	const U32 count=spOutputGroup->count();
	if(!count)
	{
		feLog("SurfaceNormalOp::handle point count is zero\n");
		catalog<String>("error")="points required";
		return;
	}

#if	FE_NOP_DEBUG
	feLog("SurfaceNormalOp::handle count %d\n",count);
#endif

	sp<SurfaceAccessorI> spPrimitiveNormal;
	if(primitiveNormals)
	{
		if(!access(spPrimitiveNormal,spSurfaceOutput,e_primitive,e_normal,
				e_quiet,e_createMissing)) return;
	}

	sp<SurfaceAccessorI> spPointNormal;
	BWORD inGroup[pointCount];
	if(pointNormals)
	{
		BWORD clearOthers=FALSE;
		if(!grouped || !access(spPointNormal,spSurfaceOutput,
				e_point,e_normal,e_quiet))
		{
			clearOthers=grouped;
			if(!access(spPointNormal,spSurfaceOutput,e_point,e_normal,
					e_quiet,e_createMissing)) return;
		}

		const SpatialVector zero(0.0,0.0,0.0);

		if(grouped)
		{
			for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
			{
				inGroup[pointIndex]=FALSE;
			}
		}

		//* clear normals
		if(clearOthers)
		{
			for(U32 index=0;index<pointCount;index++)
			{
				spPointNormal->set(index,zero);
			}
			for(U32 index=0;index<count;index++)
			{
				inGroup[spOutputGroup->integer(index)]=TRUE;
			}
		}
		else
		{
			for(U32 index=0;index<count;index++)
			{
				const I32 pointIndex=
						grouped? spOutputGroup->integer(index): index;

				inGroup[pointIndex]=TRUE;

				spPointNormal->set(pointIndex,zero);
			}
		}
	}

	//* accumulate face normals to points
	const U32 primitiveCount=spOutputVertices->count();
	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		const U32 vertexCount=spOutputVertices->subCount(primitiveIndex);

		//* TODO curves
		if(vertexCount<3)
		{
			continue;
		}

		const SpatialVector p0=
				spOutputVertices->spatialVector(primitiveIndex,0);
		const SpatialVector p1=
				spOutputVertices->spatialVector(primitiveIndex,1);
		const SpatialVector p2=
				spOutputVertices->spatialVector(primitiveIndex,2);

		const SpatialVector faceNormal= -unitSafe(cross(p1-p0,p2-p0));

#if	FE_NOP_DEBUG
	feLog("  prim %d/%d v0 %d v1 %d v2 %d\n",
				primitiveIndex,primitiveCount,
				spOutputVertices->integer(primitiveIndex,0),
				spOutputVertices->integer(primitiveIndex,1),
				spOutputVertices->integer(primitiveIndex,2));
	feLog("  p0 %s p1 %s p2 %s n %s\n",
				c_print(p0),c_print(p1),c_print(p2),c_print(faceNormal));
#endif

		if(pointNormals)
		{
			for(U32 vertexIndex=0;vertexIndex<vertexCount;vertexIndex++)
			{
				const I32 pointIndex=
						spOutputVertices->integer(primitiveIndex,vertexIndex);

				if(!grouped || inGroup[pointIndex])
				{
					const SpatialVector normal=
							spPointNormal->spatialVector(pointIndex);
					spPointNormal->set(pointIndex,normal+faceNormal);
#if	FE_NOP_DEBUG
					feLog("    vert %d/%d n %s -> %s\n",
							vertexIndex,vertexCount,
							c_print(normal),c_print(
							spPointNormal->spatialVector(pointIndex)));
#endif
				}
			}
		}

		if(primitiveNormals)
		{
			spPrimitiveNormal->set(primitiveIndex,faceNormal);
		}
	}

	//* normalize accumulated normals
	if(pointNormals)
	{
		for(U32 index=0;index<count;index++)
		{
			const I32 pointIndex=grouped? spOutputGroup->integer(index): index;

			const SpatialVector normal=
					spPointNormal->spatialVector(pointIndex);
			spPointNormal->set(pointIndex,unitSafe(normal));
		}
	}
}

} // namespace
